/* grWin1.c -
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1985, 1990 Regents of the University of California. * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  The University of California        * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.  Export of this software outside     * 
 *     * of the United States of America may require an export license.    * 
 *     *********************************************************************
 *
 * This file contains primitive functions to manipulate the Windows 32-bit
 * API.  Included here are initialization and closing
 * functions, and several utility routines used by the other Win32
 * modules.
 */

#include <stdio.h>
#ifdef	SYSV
#include <string.h>
#else
#include <strings.h>
#endif
#include <sys/types.h>
#include <signal.h>

#include <pthread.h>
#include <w32api/windows.h>
#include <w32api/windowsx.h>

#include "misc/magic.h"
#include "misc/magsgtty.h"
#include "utils/geometry.h"
#include "graphics/graphics.h"
#include "windows/windows.h"
#include "graphics/graphicsInt.h"
#include "textio/textio.h"
#include "textio/txcommands.h"
#include "signals/signals.h"
#include "utils/utils.h"
#include "utils/hash.h"
#include "grWinInt.h"
#include "misc/paths.h"

int gCmdShow;
HINSTANCE pInstance;
HDC hdc;

HWND hwnd;
HBITMAP grWinStipples[GR_NUM_STIPPLES];
HPEN grPenDraw, grPenGlyph;
HBRUSH grBrushFill, grBrushText;
LOGBRUSH grCurLogbrush;

GR_CURRENT grCurrent= {0,0,0,0,0,0};
extern unsigned int  grClass;
int grWinWStdin();
int grWinuserPC = 0;
char grWindepth[8];

HashTable	grWinWindowTable;

/* locals */

int pipeRead;
int writePipe;

pthread_mutex_t mutex;

typedef struct {
    long dashlist[8];
    int  dlen;
} LineStyle;

static LineStyle LineStyleTab[256];

#define WinDisplayHeight GrScreenRect.r_ytop
#define grMagicToWinScreen(n) (WinDisplayHeight-(n))
#define grWinScreenToMagic(n) (WinDisplayHeight-(n))

/* This is kind of a long story, and very kludgy, but the following
 * things need to be defined as externals because of the way lint
 * libraries are made by taking this module and changing all procedures
 * names "Xxxx" to "Grxxx".  The change is only done at the declaration
 * of the procedure, so we need these declarations to handle uses
 * of those names, which don't get modified.  Check out the Makefile
 * for details on this.
 */
extern Void GrWinClose(), GrWinFlush(), GrWinCreate();
extern Void GrWinDelete(),GrWinConfigure(),GrWinRaise(),GrWinLower();
extern Void GrWinLock(),GrWinUnlock(),GrWinIconUpdate();

/*---------------------------------------------------------
 * grwinSetWMandC:
 *	This is a local routine that resets the value of the current
 *	write mask and color, if necessary.
 *
 * Results:	None.
 *
 * Side Effects:    None.
 *
 * Errors:		None.
 *---------------------------------------------------------
 */

Void
grwinSetWMandC (mask, c)
    int mask;			/* New value for write mask */
    int c;			/* New value for current color */
{
    static int oldC = -1;
    static int oldM = -1;
    COLORREF locc;

    locc = grPixels[c];
    if(grCurrent.depth <= 8) {
      mask = grPlanes[mask];
      if (mask == -65) mask = -1;  /* was AllPlanes. . . */
    }
    else {
      mask = -1;  /* was AllPlanes. . . */
    }
    if (oldC == c && oldM == mask) return;

    GR_WIN_FLUSH_BATCH();

    grCurLogbrush.lbColor = locc;
    grCurLogbrush.lbStyle = BS_SOLID;

    DeletePen(grBrushText);
    grBrushText = CreateBrushIndirect(&grCurLogbrush);

    oldC = c;
    oldM = mask;
}


/*---------------------------------------------------------
 * grwinSetLineStyle:
 *	This local routine sets the current line style.
 *
 * Results:	None.
 *
 * Side Effects:
 *	A new line style is output to the display.
 *
 *---------------------------------------------------------
 */

Void
grwinSetLineStyle (style)
    int style;			/* New stipple pattern for lines. */
{
    static int oldStyle = -1;
    LineStyle *linestyle = &LineStyleTab[0];   /* default when unused */
    int winstyle;

    style &= 0xFF;
    /* if (style == oldStyle) return; */ /* must match all properties! */
    oldStyle = style;
    GR_WIN_FLUSH_BATCH();

    switch (style) {
    case 0xFF:
    case 0x00:
	winstyle = PS_SOLID;
	break;
    default:
	winstyle = PS_USERSTYLE;	/* need to account for Win95/98! */
	linestyle = &LineStyleTab[style];
	if (linestyle->dlen == 0) {

	    /* translate style to a Windows pen lpStyle */

	    long *e;
	    int offset, cnt, cur, curnew, new, i, match;

	    e = linestyle->dashlist;
	    cnt = 0;
	    offset = 1;
	    cur = 0;
	    for (i = 7; i >= 0; i--) {
		new = (style >> i) & 1;
		curnew = (cur << 1) | (long)new;
		switch (curnew) {
		case 0:
		case 3:
		    cnt++;
		    break;
		case 1:
		    if (cnt > 0) *e++ = (long)cnt; else offset = 0;
		    cnt = 1;
		    break;
		case 2:
		    *e++ = (long)cnt;
		    cnt = 1;
		    break;
		}
		cur = new;
	    }
	    *e++ = (long)cnt;
	    cnt = (int)(e - linestyle->dashlist);
	    if (offset) {
		cur = (int)e[0];
		for (i = 0; i < cnt-1; i++) e[i] = e[i+1];
		e[cnt-1] = (long)cur;
	    }
	    match = 1;
	    do {
		if (cnt % 2) break;
		for (i = 0; i < cnt/2; i++) {
		    if (e[i] != e[cnt/2 + i]) match = 0;
		}
		if (match == 0) break;
		cnt = cnt/2;
	    } while (match);
	    linestyle->dlen = cnt;
	}
    }

    grCurLogbrush.lbStyle = BS_SOLID;

    DeletePen(grPenDraw);
    grPenDraw = ExtCreatePen(winstyle, 1, &grCurLogbrush,
		linestyle->dlen, linestyle->dashlist);
}


/*---------------------------------------------------------
 * grwinSetSPattern:
 *	grwinSetSPattern associates a stipple pattern with a given
 *	stipple number.  This is a local routine called from
 *	grStyle.c .
 *
 * Results:	None.
 *
 * Side Effects:    None.
 *---------------------------------------------------------
 */

Void
grwinSetSPattern (stipple, pattern)
    int stipple;			/* The stipple number, 1-15. */
    int pattern[8];			/* 8 8-bit patterns integers */
{
    int x, y;
    unsigned char locpattern[16];

    for (y = 0; y < 8; y++)
    {
	x = y * 2;
	locpattern[x] = (unsigned char)pattern[y];
	locpattern[x + 1] = 0;  /* required for word alignment */
    }

    grWinStipples[stipple] = CreateBitmap(8, 8, 1, 1, locpattern);
}


/*---------------------------------------------------------
 * grwinSetStipple:
 *	This routine sets the current stipple number.
 *
 * Results: None.
 *
 * Side Effects:
 *	The current clipmask is set to stipple,
 *	if it wasn't that already.
 *---------------------------------------------------------
 */

Void
grwinSetStipple (stipple)
    int stipple;			/* The stipple number to be used. */
{
    static int oldStip = -1;

    /* if (stipple == oldStip) return; */ /* must match all properties! */
    oldStip = stipple;
    GR_WIN_FLUSH_BATCH();

    DeletePen(grBrushFill);

    if (stipple != 0 && stipple <= GR_NUM_STIPPLES)
    {
/*
	HBITMAP newbitmap;
	HDC aMemDC;
	HGDIOBJ tmpobj;

	aMemDC = CreateCompatibleDC(hdc);
	mapMode = GetMapMode(hdc);
	SetMapMode(aMemDC, mapMode);
	newbitmap = CreateCompatibleBitmap(hdc, 8, 8);
	tmpobj = SelectObject(aMemDC, newbitmap);
	...
	grCurLogbrush.lbHatch = (LONG)newbitmap;
*/

	grCurLogbrush.lbStyle = BS_PATTERN;
	grCurLogbrush.lbHatch = (LONG)grWinStipples[stipple];
    }
    else
	grCurLogbrush.lbStyle = BS_SOLID;

    grBrushFill = CreateBrushIndirect(&grCurLogbrush);
    fprintf(stderr, "Fill brush color = 0x%lx\n", grCurLogbrush.lbColor);
}


/*---------------------------------------------------------
 * GrWinClose:
 *
 * Results:	None.
 *
 * Side Effects:
 *---------------------------------------------------------
 */

Void
GrWinClose ()
{
    if (pInstance == NULL) return;
    TxDelete1InputDevice(pipeRead);
    close(pipeRead);
}


/*---------------------------------------------------------
 * GrWinFlush:
 * 	Flush output to display.
 *
 *	Flushing is done automatically the next time input is read,
 *	so this procedure should not be used very often.
 *
 * Results:	None.
 *
 * Side Effects:    None.
 *---------------------------------------------------------
 */

Void
GrWinFlush ()
{
    MSG locmsg;

    while(PeekMessage(&locmsg, NULL, 0, 0, PM_NOREMOVE))
    {
	if (GetMessage(&locmsg, NULL, 0, 0))
	{
	    TranslateMessage(&locmsg);
	    DispatchMessage(&locmsg);
	}
	else
	{
	    MainExit(0);
	}
    }
}


/*
 * ---------------------------------------------------------------------------
 *
 * grWinStdin --
 *
 *      Handle the stdin device for the Windows driver.
 *	(Key and Button Events only)
 *
 * Results:
 *      None.
 *
 * Side Effects:
 *      Adds events to the data queue.
 *
 * ---------------------------------------------------------------------------
 */

Void
grWinStdin ()
{
    TxInputEvent *event;
    HashEntry	*entry;

    UINT	iMsg;
    HWND	locwind;
    WPARAM	wParam;
    LPARAM	lParam;
    int		c;

    read(pipeRead, &locwind, sizeof(HWND));
    read(pipeRead, &iMsg, sizeof(unsigned int));

    /* Remainder of pipe read is context-dependent */

    if (iMsg == WM_KEYDOWN || iMsg == WM_SYSKEYDOWN || iMsg == WM_CHAR)
    {
	read(pipeRead, &c, sizeof(int));
    }
    else
    {
	read(pipeRead, &wParam, sizeof(WPARAM));
	read(pipeRead, &lParam, sizeof(LPARAM));
    }

    switch (iMsg) 
    {
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
	case WM_MBUTTONDOWN:
	case WM_MBUTTONUP:
	    {
	        event = TxNewEvent();
		switch (iMsg) {
		case WM_LBUTTONDOWN: case WM_LBUTTONUP:
		    event->txe_button = TX_LEFT_BUTTON;
		    break;
		case WM_MBUTTONDOWN: case WM_MBUTTONUP:
		    event->txe_button = TX_MIDDLE_BUTTON;
		    break;
		case WM_RBUTTONDOWN: case WM_RBUTTONUP:
		    event->txe_button = TX_RIGHT_BUTTON;
		    break;
		}
		switch(iMsg) {
		case WM_LBUTTONUP: case WM_RBUTTONUP: case WM_MBUTTONUP:
		    event->txe_buttonAction = TX_BUTTON_UP;
		    break;
		case WM_LBUTTONDOWN: case WM_RBUTTONDOWN: case WM_MBUTTONDOWN:
		    event->txe_buttonAction = TX_BUTTON_DOWN;
		    break;
		}

	        grCurrent.window = locwind;
		entry = HashLookOnly(&grWinWindowTable,grCurrent.window);
	        grCurrent.mw= (entry)?(MagWindow *)HashGetValue(entry):0;

		event->txe_p.p_x = LOWORD(lParam);
		event->txe_p.p_y = grWinToMagic(HIWORD(lParam));
		event->txe_wid = grCurrent.mw->w_wid;
		TxAddEvent(event);
	    }
	    break;
	case WM_KEYDOWN: case WM_SYSKEYDOWN: case WM_CHAR:
	    {
	        event = TxNewEvent();

	        grCurrent.window = locwind;
		entry = HashLookOnly(&grWinWindowTable,grCurrent.window);
	        grCurrent.mw= (entry)?(MagWindow *)HashGetValue(entry):0;

		if (c == (int)'\r') c = (int)'\n';
		event->txe_button = TX_CHARACTER;
		event->txe_ch = c;
		event->txe_buttonAction = TX_KEY_DOWN;
		grwinGetCursorPos(&(event->txe_p));
		event->txe_wid = grCurrent.mw->w_wid;
		TxAddEvent(event);
	    } 
	    break;
	default:
	    TxPrintf("Event %d should not be processed here.\n", iMsg);
     }
}

/*----------------------------------------------------------------------*/
/* Initialize the mutex locking mechanism				*/
/*----------------------------------------------------------------------*/

void
initialize_lock()
{
   pthread_mutexattr_t mutex_attrib;

   pthread_mutexattr_init(&mutex_attrib);
   pthread_mutexattr_settype(&mutex_attrib, PTHREAD_MUTEX_RECURSIVE);
   pthread_mutex_init(&mutex, &mutex_attrib);
}
  
/*----------------------------------------------------------------------*/
/* Wrapper functions for pthread's mutex lock and mutex functions       */
/*----------------------------------------------------------------------*/

void
acquire_lock()
{ 
   pthread_mutex_lock(&mutex);
} 
  
/*----------------------------------------------------------------------*/

void
release_lock()
{ 
   pthread_mutex_unlock(&mutex);
} 

/*---------------------------------------------------------
 * winloop_begin:  Threaded version of XHelper7 (X11Helper.c)
 *
 *---------------------------------------------------------
 */

void
winloop_begin()
{
    MSG msg;
    WNDCLASSEX wndclass;
    static char szClassName[] = "Magic";
    int bpp;

    HashInit(&grWinWindowTable,8,HT_WORDKEYS);

    wndclass.cbSize = sizeof(WNDCLASSEX);
    wndclass.style = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc = ParseMessage;
    wndclass.cbClsExtra = 0;
    wndclass.cbWndExtra = 0;
    wndclass.hInstance = pInstance;
    wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wndclass.lpszMenuName = NULL;
    wndclass.lpszClassName = szClassName;
    wndclass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

    RegisterClassEx(&wndclass);

    hwnd = CreateWindowEx(
		WS_EX_LEFT,		// Left aligned (default)
		szClassName,		// pointer to class name
		"WinMagic",		// window caption
		WS_OVERLAPPEDWINDOW,	// window style
		CW_USEDEFAULT,		// initial x position
		CW_USEDEFAULT, 		// initial y position
		CW_USEDEFAULT,		// initial x size
		CW_USEDEFAULT,		// initial y size
		NULL,			// parent window handle
		NULL,			// window menu handle
		pInstance,		// program instance
		NULL);			// creation parameters

    hdc = GetDC(hwnd);
    fprintf(stderr, "Window created with hwnd=0x%x, hdc=0x%x\n",
	hwnd, hdc);

    GrScreenRect.r_xtop = GetDeviceCaps(hdc, HORZRES);
    GrScreenRect.r_ytop = GetDeviceCaps(hdc, VERTRES);
    fprintf(stderr, "Screen dimensions are 0,0 to %d, %d\n",
	GrScreenRect.r_xtop, GrScreenRect.r_ytop);

    /* Get the properties of the Windows display */

    bpp = GetDeviceCaps(hdc, BITSPIXEL);
    fprintf(stderr, "Reported bits per pixel is %d\n", bpp);
    if (bpp <= 8)
    {
	grCurrent.depth = GetDeviceCaps(hdc, PLANES);
    }
    else
	grCurrent.depth = bpp;

    fprintf(stderr, "Display depth is %d\n", grCurrent.depth);

    if (grWinuserPC > 0)
    {
    	if (grWinuserPC > grCurrent.depth)
	{
	    TxError(
	 "You cannot specify more planes (%d) than the display has (%d).\n",
			grWinuserPC,grCurrent.depth);
	    MainExit(1);
	}
	grCurrent.depth = (grWinuserPC == 1)?0:grWinuserPC;
    }
    else  /* default to nearest depth that we "should" have dstyles and 
    	   * colormap for.
	   */
    {
        if      (grCurrent.depth < 4) grCurrent.depth = 0;
        else if (grCurrent.depth < 6) grCurrent.depth = 4;
        else if (grCurrent.depth < 7) grCurrent.depth = 6;
	else if (grCurrent.depth <= 8) grCurrent.depth = 7;
    }
    grDStyleType = grWindepth;
    if (grCurrent.depth) 
    {
	sprintf(grWindepth,"%dbit",grCurrent.depth);
	grCMapType = grWindepth;
    }
    else
    {
    	sprintf(grWindepth,"bw");
        grCMapType = NULL;
	/* have to call GrWinSetCmap directly here; with colormap
	 * set to NULL, it doesn't get called otherwise.
         */
        GrWinSetCMap((char *)NULL);
    }

    ShowWindow(hwnd, gCmdShow);
    UpdateWindow(hwnd);

    while (GetMessage(&msg, hwnd, 0, 0)) {
        /* fprintf(stderr, "WinLoop got message %d\n", msg); */
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }
    fprintf(stderr, "WinLoop ended\n");
}  


/*---------------------------------------------------------
 * winSetDisplay:
 *	This routine sets the appropriate parameters so that
 *	Magic will work with the Windows display.
 *
 * Results:  success / fail
 *
 * Side Effects:	Sets up the pipe.
 *---------------------------------------------------------
 */

bool
winSetDisplay (dispType, outFileName, mouseFileName)
    char *dispType;
    char *outFileName;
    char *mouseFileName;
{
    int	 fildes[2], fildes2[2];
    char *planecount;
    char *fullname;
    FILE *f;
    bool execFailed = false;
    pthread_t thread;

    WindPackageType = WIND_MS_WINDOWS;

    grCursorType = "bw";
    
    if ((int)strlen(dispType) > 5)  /* 5 == strlen("XWIND") */
    {
    	 planecount = dispType+5;
	 grWinuserPC = atoi(planecount);
    }
    
    WindScrollBarWidth = 14;

    /* Set up helper process */
    pipe(fildes);
    pipeRead = fildes[0];
    writePipe = fildes[1];

    TxAdd1InputDevice(pipeRead, grWinStdin, (ClientData) NULL);

    /* Set up the procedure values in the indirection table. */

    GrInitPtr = NULL;
    GrLockPtr = GrWinLock;
    GrUnlockPtr = GrWinUnlock;
    GrClosePtr = GrWinClose;
    GrSetCMapPtr = GrWinSetCMap;

    GrEnableTabletPtr = GrWinEnableTablet;
    GrDisableTabletPtr = GrWinDisableTablet;
    GrSetCursorPtr = GrWinSetCursor;
    GrTextSizePtr = GrWinTextSize;
    GrDrawGlyphPtr = GrWinDrawGlyph;
    GrReadPixelPtr = GrWinReadPixel;
    GrFlushPtr = GrWinFlush;

    GrCreateWindowPtr = GrWinCreate;
    GrDeleteWindowPtr = GrWinDelete;
    GrConfigureWindowPtr = GrWinConfigure;
    GrOverWindowPtr = GrWinRaise;
    GrUnderWindowPtr = GrWinLower;
    GrUpdateIconPtr = GrWinIconUpdate; 

    /* local indirections */
    grSetSPatternPtr = grwinSetSPattern;
    grPutTextPtr = grwinPutText;
    grDefineCursorPtr = grwinDefineCursor;
    GrBitBltPtr = GrWinBitBlt;
    grDrawGridPtr = grwinDrawGrid;
    grDrawLinePtr = grwinDrawLine;
    grSetWMandCPtr = grwinSetWMandC;
    grFillRectPtr = grwinFillRect;
    grSetStipplePtr = grwinSetStipple;
    grSetLineStylePtr = grwinSetLineStyle;
    grSetCharSizePtr = grwinSetCharSize;
    grMaxStipples = 32;
#ifdef NONMANHATTAN
    grFillPolygonPtr = grwinFillPolygon;
#endif
    
    if (execFailed) {
	TxError("Execution failed!\n");
	return false;
    }

    TxAdd1InputDevice(fileno(stdin), grWinWStdin, (ClientData) NULL);

    initialize_lock();

    /* This thread acquires the lock and won't release it until */
    /* the first window has been created and registered in the	*/
    /* hash table.						*/

    acquire_lock();

    pthread_create(&thread, NULL, (void *)winloop_begin, NULL);

    return true;
}

/*
 * ----------------------------------------------------------------------------
 *
 * grWinWStdin --
 *      Handle the stdin device for the Windows interface.
 *
 * Results:
 *      None.
 *
 * Side Effects:
 *      Adds events to the event queue.
 *
 * ----------------------------------------------------------------------------
 */
 /*ARGSUSED*/

Void
grWinWStdin(fd, cdata)
    int fd;
    ClientData cdata;
{
    int ch;
    TxInputEvent *event;

    event = TxNewEvent();
    ch = getc(stdin);
    if (ch == EOF)
	event->txe_button = TX_EOF;
    else
	event->txe_button = TX_CHARACTER;
    event->txe_ch = ch;
    event->txe_buttonAction = 0;
    event->txe_wid = WIND_NO_WINDOW;
    event->txe_p.p_x = GR_CURSOR_X;
    event->txe_p.p_y = GR_CURSOR_Y;
    TxAddEvent(event);
}

/*------------------------------------------------------*/
/* ParseMessage is copied from X11Helper.c		*/
/*------------------------------------------------------*/

LRESULT CALLBACK ParseMessage (lwind, iMsg, wParam, lParam)
    HWND  lwind;
    UINT iMsg;
    WPARAM wParam;
    LPARAM lParam;
{
    int keywstate;	/* KeySym with prepended state information */
    int keymod;
    PAINTSTRUCT ps;
    HashEntry	*entry;
    LRESULT status;

    /* acquire_lock(); */
    switch (iMsg)
    {
	/* case WM_KEYDOWN: case WM_SYSKEYDOWN: */ /* no? */
	case WM_CHAR:
    	    /* fprintf(stderr, "Processing key message %d\n", iMsg); */

	    /* These keymods correspond to X11 types, and are translated */
	    /* by the MacroName command.				     */

	    keymod = 0;
	    if (GetKeyState(VK_SHIFT) < 0) keymod |= 0x1;
	    if (GetKeyState(VK_NUMLOCK) != 0) keymod |= 0x2;
	    if (GetKeyState(VK_CONTROL) < 0) keymod |= 0x4;
	    if (GetKeyState(VK_MENU) < 0) keymod |= 0x8;
	    /* fprintf(stderr, "keymod is 0x%x\n", keymod); */

	    if (iMsg != WM_CHAR)  	/* No ASCII equivalent to string */
	    {
	        keywstate = (keymod << 16) | (wParam & 0xffff);

	        write(writePipe, &lwind, sizeof(HWND));
	        write(writePipe, &iMsg, sizeof(UINT));
	        write(writePipe, &keywstate, sizeof(int));
		/* fprintf(stderr, "keywstate is 0x%x as WM_KEYDOWN\n",
			keywstate); */
	    }
	    else			/* For keys with ASCII equivalent */
	    {
	        /* If Control or Shift is used alone, it should be removed */
		/* from the modifier list, since the ASCII value subsumes */
		/* the Control and Shift functions.  However, Control +	*/
		/* Shift + key should be retained, as it cannot be	*/
		/* expressed in ASCII.					*/

	        int c;

	        if (!(keymod & (0x2 | 0x8))) {
		    if (!(keymod & 0x4))
		        keymod &= ~0x1;
		    else if (!(keymod & 0x1))
		        keymod &= ~0x4;
	        }

	        c = (int)wParam;
	        if (c == 3)		/* Ctrl-C interrupt */
	        {
		    int parentID = (int)getppid();
		    kill(parentID, SIGINT);
	        } 
	        else
	        {
		    /* When Control modifier is present, use the capital */
		    /* letter value instead of the control value.	 */

		    if ((keymod & 0x4) && (c < 32))
		        c += 'A' - 1;

		    keywstate = (keymod << 16) | ((int)c & 0xff);

		    write(writePipe, &lwind, sizeof(HWND));
		    write(writePipe, &iMsg, sizeof(UINT));
		    write(writePipe, &keywstate, sizeof(int));
		    /* fprintf(stderr, "keywstate is 0x%x as WM_CHAR\n",
			keywstate); */
	        }
	    }
	    break;
	case WM_LBUTTONDOWN: case WM_LBUTTONUP: case WM_RBUTTONDOWN:
	case WM_RBUTTONUP: case WM_MBUTTONDOWN: case WM_MBUTTONUP:
    	    /* fprintf(stderr, "Processing button message %d\n", iMsg); */
	    write(writePipe, &lwind, sizeof(HWND));
	    write(writePipe, &iMsg, sizeof(UINT));
	    write(writePipe, &wParam, sizeof(WPARAM));
	    write(writePipe, &lParam, sizeof(LPARAM));
	    break;

	case WM_PAINT:
	    {
		Rect screenRect;
		RECT winrect;
		MagWindow	*w;
		    
    	        /* fprintf(stderr, "Processing paint message %d\n", iMsg); */
	        grCurrent.window = lwind;
		entry = HashLookOnly(&grWinWindowTable,grCurrent.window);
	        w = (entry)?(MagWindow *)HashGetValue(entry):0;
		if (!w)
		{
		    fprintf(stderr, "WM_PAINT: waiting on thread\n");
		    acquire_lock();	/* Wait on other thread */
		    release_lock();
		    entry = HashLookOnly(&grWinWindowTable,grCurrent.window);
	            w = (entry)?(MagWindow *)HashGetValue(entry):0;
		}
		if (w)
		{
	            grCurrent.mw=w;

		    hdc = BeginPaint(lwind, &ps);
		    GetClientRect(lwind, &winrect);
		    acquire_lock();
		    screenRect.r_xbot = winrect.left;
            	    screenRect.r_xtop = winrect.right;
            	    screenRect.r_ytop = grWinToMagic(winrect.top);
            	    screenRect.r_ybot = grWinToMagic(winrect.bottom);
		    fprintf(stderr, "Painting area (%d, %d) to (%d, %d)\n",
			screenRect.r_xbot, screenRect.r_ybot,
			screenRect.r_xtop, screenRect.r_ytop);
		    fprintf(stderr, "Windows area (%d, %d) to (%d, %d)\n",
			winrect.left, winrect.bottom, winrect.right,
			winrect.top);
                    WindAreaChanged(w, &screenRect);
                    WindUpdate();
		    release_lock();
		    EndPaint(lwind, &ps);
		    fprintf(stderr, "End Paint\n");
		}
		else
		    fprintf(stderr, "WM_PAINT: lwind = 0x%x, w is NULL\n",
				lwind);
            }
	    break;
	case WM_SIZE:
	    {
		Rect screenRect;
		MagWindow	*w;
		RECT winrect;
		    
    	        /* fprintf(stderr, "Processing size message %d\n", iMsg); */
	        grCurrent.window = lwind;
		entry = HashLookOnly(&grWinWindowTable,grCurrent.window);
	        w = (entry)?(MagWindow *)HashGetValue(entry):0;
	        grCurrent.mw=w;

		if (w)
		{
		    hdc = BeginPaint(lwind, &ps);
		    GetClientRect(lwind, &winrect);
		    acquire_lock();
		    screenRect.r_xbot = winrect.left;
            	    screenRect.r_xtop = winrect.right;
            	    screenRect.r_ytop = grWinScreenToMagic(winrect.bottom);
            	    screenRect.r_ybot = grWinScreenToMagic(winrect.top);

		    WindReframe(w,&screenRect,FALSE,FALSE);
		    WindRedisplay(w);
		    release_lock();
		    EndPaint(lwind, &ps);
		}
		else
		{
		    fprintf(stderr, "WM_SIZE: w is NULL\n");
		}
            }
            break;
	case WM_CREATE:
	    {
		MagWindow	*w;
		    
    	        /* fprintf(stderr, "Processing create message %d\n", iMsg); */
	        grCurrent.window = lwind;
		entry = HashLookOnly(&grWinWindowTable, grCurrent.window);
	        w = (entry)?(MagWindow *)HashGetValue(entry):0;

                if (w)
		{
		    SigDisableInterrupts();
	    	    WindView(w);
		    SigEnableInterrupts();
		}
		else
		    fprintf(stderr, "WM_CREATE: w is NULL\n");
	    }
            break;
	case WM_DESTROY:
    	    /* fprintf(stderr, "Processing destroy message %d\n", iMsg); */
	    PostQuitMessage(0);
	    break;
	default:
    	    /* fprintf(stderr, "Processing unknown message %d\n", iMsg); */
	    status = DefWindowProc(lwind, iMsg, wParam, lParam);
	    /* release_lock(); */
	    return status;
    }
    /* release_lock(); */
    return 0;
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinCreate --
 *      Create a new window under Win32API.
 *	Bind Windows window to Magic Window w.
 *
 * Results:
 *      Window created, window ID sent to winloop thread.
 *
 * Side Effects:
 *	None.
 *
 * ----------------------------------------------------------------------------
 */

int
GrWinCreate(w)
    MagWindow *w;
{
    static int firstWindow = 1;
    HashEntry	*entry;
    char	*option = (firstWindow)?"window":"newwindow";

    WindSeparateRedisplay(w);


    if (!firstWindow)
    {
	/* Create a new window here for all windows except the 1st */
	/* Return a new hwnd for the new window.		   */
    }

    grCurrent.window = hwnd; 
    grCurrent.mw = w;
    w->w_grdata = (ClientData) hwnd;
	
    entry = HashFind(&grWinWindowTable, grCurrent.window);
    HashSetValue(entry, w);
    fprintf(stderr, "Setting hash entry for window 0x%x\n", hwnd);


    /* Must release the mutex lock after doing HashSetValue(),	*/
    /* but *only* on the first window.				*/

    if (firstWindow)
    {
	RECT windrect;

	/* The first window was created outside this routine, so */
	/* we must query its size and position.			 */

	GetWindowRect(hwnd, &windrect);

	w->w_frameArea.r_xbot = 0;
	w->w_frameArea.r_ybot = 0;
	w->w_frameArea.r_xtop = windrect.right - windrect.left;
	w->w_frameArea.r_ytop = windrect.bottom - windrect.top;

	/* shrink sides to compensate for border. . . ?? */
	w->w_frameArea.r_xbot += 5;
	w->w_frameArea.r_ybot += 5;
	w->w_frameArea.r_xtop -= 5;
	w->w_frameArea.r_ytop -= 20;
	/* End temporary code */

	WindReframe(w, &(w->w_frameArea), FALSE, FALSE);

	grCurLogbrush.lbStyle = BS_SOLID;
	grCurLogbrush.lbColor = 0;
	grCurLogbrush.lbHatch = 0L;

        grBrushFill = CreateBrushIndirect(&grCurLogbrush);
	grBrushText = CreateBrushIndirect(&grCurLogbrush);

        grPenDraw = CreatePen(
			PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_MITER,
			1, 0);
        grPenGlyph = CreatePen(
			PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_MITER,
			1, 0);
	grwinLoadFont();
    }

    SetCursor(grCurrent.cursor);
    GrWinIconUpdate(w,w->w_caption); 

    if (firstWindow)
    {
	release_lock();		/* Let the event loop thread run */
				/* See main/main.c . . . */
	firstWindow = 0;
    }

    return 1;
}


/*
 * ----------------------------------------------------------------------------
 *
 * GrWinDelete --
 *      Destroy a window.
 *
 * Results:
 *      Window destroyed.
 *
 *
 * ----------------------------------------------------------------------------
 */

GrWinDelete(w)
    MagWindow *w;
{
    HWND lwin;
    HashEntry	*entry;

    lwin = (HWND) w->w_grdata;
    entry = HashLookOnly(&grWinWindowTable,lwin);
    HashSetValue(entry,NULL);
    
    DestroyWindow(lwin);
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinConfigure --
 *      Resize/ Move an existing window.
 *
 * Results:
 *      Window reconfigured to w->w_frameArea.
 *
 * Side Effects:
 *      None.
 *
 * ----------------------------------------------------------------------------
 */
 /*ARGSUSED*/

GrWinConfigure(w)
    MagWindow *w;
{
    MoveWindow((HWND)w->w_grdata,
	    w->w_frameArea.r_xbot,
	    grMagicToWinScreen(w->w_frameArea.r_ybot),
	    w->w_frameArea.r_xtop - w->w_frameArea.r_xbot,
	    w->w_frameArea.r_ytop - w->w_frameArea.r_ybot,
	    true);
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinRaise --
 *      Raise a window to the top of the screen such that nothing
 *	obscures it.
 *
 * Results:
 *      Window raised.
 *
 * Side Effects:
 *      None.
 *
 * ----------------------------------------------------------------------------
 */

GrWinRaise(w)
    MagWindow *w;
{
    /* What is this function in Windows? */
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinLower --
 *      Lower a window below all other windows.
 *	obscures it.
 *
 * Results:
 *      Window lowered.
 *
 * Side Effects:
 *      None.
 *
 * ----------------------------------------------------------------------------
 */

GrWinLower(w)
    MagWindow *w;
{
    /* What is this function in Windows? */
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinLock --
 *      Lock a window and set global variables "grCurrent.window"
 *	and "grCurrent.mw" to reference the locked window.
 *
 * Results:
 *      Window locked.
 *
 * Side Effects:
 *	None.
 *
 * ----------------------------------------------------------------------------
 */

GrWinLock(w, flag)
    MagWindow *w;
    bool flag;
{
    acquire_lock();
    grSimpleLock(w, flag);
    if ( w != (MagWindow *) GR_LOCK_SCREEN )
    {
	grCurrent.mw = w;
	grCurrent.window = (HWND) w->w_grdata;
    }
}

/*
 * ----------------------------------------------------------------------------
 *
 * GrWinUnlock --
 *      Unlock a window, flushing stuff out to the display.
 *
 * Results:
 *      Window unlocked.
 *
 * Side Effects:
 *	Display update.
 *
 * ----------------------------------------------------------------------------
 */

GrWinUnlock(w)
    MagWindow *w;
{
    GR_WIN_FLUSH_BATCH();
    grSimpleUnlock(w);
    release_lock();
}

/*
 *-------------------------------------------------------------------------
 *
 * GrWinIconUpdate -- updates the icon text with the window script
 *
 * Results: none
 *
 * Side Effects: changes the icon text
 *
 *-------------------------------------------------------------------------
 */
Void GrWinIconUpdate(w,text)
    MagWindow	*w;
    char	*text;

{
     HWND	wind = (HWND)(w->w_grdata);
     char	*brack;
     
     if (w->w_grdata == NULL) return;
     if (brack = index(text,'['))
     {
     	  brack--;
	  *brack = 0;
	  /* What is the Windows equivalent?	*/
          /* XSetIconName(grXdpy,wind,text);	*/
	  /* XStoreName(grXdpy,wind,text);	*/
     	  *brack = ' ';
	  return;
     }
     if (brack = rindex(text,' ')) text = brack+1;
     /* What is the Windows equivalent?	*/
     /* XSetIconName(grXdpy,wind,text); */
     /* XStoreName(grXdpy,wind,text);	*/
}

