/* grWin2.c -
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1985, 1990 Regents of the University of California. * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  The University of California        * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.  Export of this software outside     * 
 *     * of the United States of America may require an export license.    * 
 *     *********************************************************************
 *
 * This file contains additional functions to manipulate an X
 * color display.  Included here are rectangle drawing and color map
 * loading.
 */

#include <stdio.h>
#include <w32api/windows.h>
#include <w32api/windowsx.h>

#include "misc/magic.h"
#include "textio/textio.h"
#include "utils/geometry.h"
#include "graphics/glyphs.h"
#include "graphics/graphics.h"
#include "windows/windows.h"
#include "graphics/graphicsInt.h"
#include "grWinInt.h"

extern char *DBWStyleType;
unsigned long grPlanes[256];	/* Unused for 16,24-bit color */
COLORREF grPixels[256];		/* Unique colors used by Magic */

/* Unused for now;  these are for the 8-bit color display type */
#define PALETTE_BASE		128
#define PALETTE_RESERVED	0


/*---------------------------------------------------------
 * GrWinSetCMap:
 *	Set the colormap.  Fill it if it is an 8-bit indexed
 *	map.
 *
 * Results:	None.
 *
 * Side Effects:
 *	The values in the color map are set from the array indicated
 *	by pmap.
 *
 * Errors:		None.
 *
 *---------------------------------------------------------
 */

Void
GrWinSetCMap (pmap)
    char *pmap;			/* A pointer to 256*3 bytes containing the
				 * new values for the color map.  The first
				 * three values are red, green, and blue
				 * intensities for color 0, and so on.
				 */
{
    char *p;	
    int i;

    static int firstCall = 1;
    static int planeCount;	/* how many planes we want to allocate */
    static int colorCount;	/* how many colors we want to allocate */

    fprintf(stderr, "GrWinSetCMap() 1\n");
    if (firstCall)
    {
	firstCall = 0;

        fprintf(stderr, "GrWinSetCMap() 2\n");
	planeCount = grCurrent.depth;
	colorCount = 256;

	if (planeCount > 8) /* must load appropriate Dstyle file. */
	{
	    grDStyleType = "24bit";
	    if (GrLoadStyles(DBWStyleType, ".", SysLibPath))  MainExit();
	}

	if (planeCount == 8)
	{
            grCurrent.depth = 7;
            planeCount = 7;      /* This resets to old 7-plane mode */
            colorCount = 1 << (planeCount);
	}
    }

    /* Need to instantiate proper handling of 8-bit graphics mode */

    fprintf(stderr, "GrWinSetCMap() 3\n");
    if (grCurrent.depth)
    {
	p = pmap;
	for (i = 0; i < colorCount; i++) 
	{
	    int r, g, b;
	    r = (int)*p++;
	    g = (int)*p++;
	    b = (int)*p++;
	    /* grPixels[i] = RGB(*p++, *p++, *p++); */
	    grPixels[i] = RGB(r, g, b);
	    fprintf(stderr, "Color %d is 0x%lx\n", i, grPixels[i]);
	}
    }
}

/*------------------------------------------------------*/ 
/* Windows can't batch rectangles efficiently, but it	*/
/* can handle lines in batch mode with Polyline.	*/
/*------------------------------------------------------*/ 

POINT grwinLines[WIN_BATCH_SIZE];
int grwinNbLines=0;

/*---------------------------------------------------------
 * grwinDrawLines:
 *	This routine draws a batch of lines.
 *
 * Results:	None.
 *
 * Side Effects:
 *	Draw a bunch of lines.
 *---------------------------------------------------------
 */
Void
grwinDrawLines(lines, nb)
    POINT lines[];
    int nb;
{
    SelectPen(hdc, grPenDraw);
    Polyline(hdc, lines, nb);
}


/*---------------------------------------------------------
 * grwinDrawLine:
 *	This routine draws a line.
 *
 * Results:	None.
 *
 * Side Effects:
 *	Draw a line for (x1, y1) to (x2, y2) inclusive.
 *---------------------------------------------------------
 */
Void
grwinDrawLine (x1, y1, x2, y2)
    int x1, y1;			/* Screen coordinates of first point. */
    int x2, y2;			/* Screen coordinates of second point. */
{
    if (grwinNbLines == WIN_BATCH_SIZE) GR_WIN_FLUSH_LINES();
    grwinLines[grwinNbLines].x = x1;
    grwinLines[grwinNbLines].y = grMagicToWin(y1);
    grwinLines[grwinNbLines + 1].x = x2;
    grwinLines[grwinNbLines + 1].y = grMagicToWin(y2);
    grwinNbLines += 2;
}

/*---------------------------------------------------------
 * grwinFillRect:
 *	This routine draws a solid rectangle.
 *
 * Results:	None.
 *
 * Side Effects:
 *	Drawing.
 *---------------------------------------------------------
 */

Void
grwinFillRect(r)
    register Rect *r;	/* Address of a rectangle in screen
			 * coordinates.
			 */
{
    RECT winrect;

    grwinRectConvert(r, &winrect);
    /* SelectBrush(hdc, grBrushFill); */ /* Necessary or not? */

    /* May want to do regions instead, and batch them? */
    FillRect(hdc, &winrect, grBrushFill);

    /* Rectangle(hdc, r->r_xbot, grMagicToWin(r->r_ytop),
		r->r_xtop, grMagicToWin(r->r_ybot)); */

    fprintf(stderr, "grwinFillRect:  dimensions are (%d, %d) to (%d, %d)\n",
	r->r_xbot, grMagicToWin(r->r_ytop),
	r->r_xtop, grMagicToWin(r->r_ybot));
	
}

#ifdef NONMANHATTAN

/*---------------------------------------------------------
 * grwinFillPolygon:
 *	This routine draws a solid polygon
 *
 * Results:     None.
 *
 * Side Effects:
 *	Drawing.
 *---------------------------------------------------------
 */

Void
grwinFillPolygon(tp, np)
    Point *tp;
    int np;
{
    POINT xp[5];
    int i;

    for (i = 0; i < np; i++)
    {
	xp[i].x = tp[i].p_x;
	xp[i].y = grMagicToWin(tp[i].p_y);
    }
    Polygon(hdc, xp, np);
}

#endif NONMANHATTAN
